// ---------------------------------------------------------------------------
//	M88 - PC-88 Emulator
//	Copyright (C) cisc 1998.
// ---------------------------------------------------------------------------
//	Emulation of Floppy Disk Controller uPD765A
// ---------------------------------------------------------------------------
//	$Id: FDC.cpp,v 1.2 1999/01/11 15:32:08 cisc Exp $

#include <stdio.h>
#include <string.h>
#include "FDC.h"
#include "FDU.h"
#include "SubSys.h"

#ifndef RELEASEBUILD
	#define FDC_LOGGING
#endif

#ifdef FDC_LOGGING
	static FILE* fp = 0;

	static void LogBegin()   { if (!fp) fp = fopen("fdc.log", "w"); }
	static void LogEnd()     { if (fp) fclose(fp); fp = 0; }
	#define Log0(a)          { if (fp) fprintf(fp, a); }
	#define Log1(a, b)       { if (fp) fprintf(fp, a, b); }
	#define Log2(a, b, c)    { if (fp) fprintf(fp, a, b, c); }
	#define Log3(a, b, c, d) { if (fp) fprintf(fp, a, b, c, d); }
	#define Log4(a, b,c,d,e) { if (fp) fprintf(fp, a, b, c, d, e); }
#else
	#define LogBegin()		 (void) 0
	#define LogEnd()		 (void) 0
	#define Log0(a)			 (void) 0
	#define Log1(a, b)		 (void) 0
	#define Log2(a, b, c)	 (void) 0
	#define Log3(a, b, c, d) (void) 0
	#define Log4(a, b,c,d,e) (void) 0
#endif

// ---------------------------------------------------------------------------
//	\z/
// ---------------------------------------------------------------------------

FDC::FDC() 
{
	LogBegin();
	for (int i=0; i<NUM_FLOPPY_DRIVES; i++)
		fdu[i] = 0;
}

FDC::~FDC()
{
	delete[] buffer;
	
	for (int i=0; i<NUM_FLOPPY_DRIVES; i++)
		delete fdu[i];
	
	LogEnd();
}

// ---------------------------------------------------------------------------
//	
// ---------------------------------------------------------------------------

bool FDC::Initialize(SubSystem* sub)
{
	system = sub;
	seekstate = 0;
	buffer = new uint8[0x4000];
	if (!buffer) return false;
	return true;
}

// ---------------------------------------------------------------------------
//	Reset
// ---------------------------------------------------------------------------

void FDC::Reset()
{
	Log0("Reset!\n");
	ShiftToIdlePhase();
	int_requested = false;
	system->IntrReset();
}

// ---------------------------------------------------------------------------
//	FDC::Status
//	Xe[^XWX^
//
//	MSB                         LSB
//	RQM DIO NDM CB  D3B D2B D1B D0B
//
//	CB  = IdlePhase ȊO
//	NDM = E-Phase
//	DIO = 0 Ȃ CPU->FDC (Put)  1 Ȃ FDC->CPU (Get)
//	RQM = f[^̑MEM̗pӂł
// ---------------------------------------------------------------------------

uint8 FDC::Status()
{
	return seekstate | status;
}

// ---------------------------------------------------------------------------
//	Phase ڍs
// ---------------------------------------------------------------------------

bool FDC::ShiftPhase(Phase ph, uint8 stbits, bool notc)
{
	if (notc && tcin)
		return false;
	phase = ph;
	status = stbits;
	return true;
}

void FDC::ShiftToIdlePhase()
{
	ShiftPhase(IdlePhase, S_RQM, false);
	tcin = false;
}

void FDC::ShiftToCommandPhase(int nbytes)
{
	ShiftPhase(CommandPhase, S_RQM | S_CB, false);
	status = S_RQM | S_CB;
	bufptr = buffer;
	count = nbytes;
	int_requested = false;
	system->IntrReset();
}

void FDC::ShiftToExecutePhase(uint8 direction)
{
	if (ShiftPhase(ExecutePhase, direction | S_RQM | S_NDM | S_CB, true))
	{
		system->Intr();
	}
}

void FDC::ShiftToExecReadPhase(int nbytes)
{
	if (ShiftPhase(ExecReadPhase, S_RQM | S_DIO | S_NDM | S_CB, true))
	{
		bufptr = buffer;
		count = nbytes;
	}
}

void FDC::ShiftToExecWritePhase(int nbytes)
{
	if (ShiftPhase(ExecWritePhase, S_RQM | S_NDM | S_CB, true))
	{
		bufptr = buffer;
		count = nbytes;
	}
}

void FDC::ShiftToErrorPhase(uint8 dir)
{
	ShiftPhase(ErrorPhase, S_RQM | S_CB | S_NDM | dir, true);
	system->Intr();
	count = -1;
}


void FDC::ShiftToResultPhase(int nbytes)
{
	ShiftPhase(ResultPhase, S_RQM | S_CB | S_DIO, false);
	bufptr = buffer;
	count = nbytes;
	Log0("\t{");
}

// common result phase (ST0/ST1/ST2/C/H/R/N)

void FDC::ShiftToResultPhase7()
{
	buffer[0] = (result & 0xf8) | (hdu & 7);
	buffer[1] = uint8(result >>  8);
	buffer[2] = uint8(result >> 16);
	buffer[3] = idr.c;
	buffer[4] = idr.h;
	buffer[5] = idr.r;
	buffer[6] = idr.n;
	system->Intr();
	ShiftToResultPhase(7);
}

// ---------------------------------------------------------------------------
//	FDC::SetData
//	CPU  FDC Ƀf[^𑗂
// ---------------------------------------------------------------------------

inline void FDC::SetDataIdle()
{
	Log1("[%.2x] ", data);
	command = data;
	(this->*CommandTable[command & 31])();
}

inline void FDC::SetDataCommand()
{
	*bufptr++ = data;
	if (--count)
	{
		status |= S_RQM;
	}
	else
	{
		(this->*CommandTable[command & 31])();
	}
}

inline void FDC::SetDataExecute()
{
	(this->*CommandTable[command & 31])();
	if (phase == ExecWritePhase)
		SetDataExecWrite();
}

inline void FDC::SetDataExecWrite()
{
	*bufptr++ = data;
	if (--count)
	{
		system->Intr();
		status |= S_RQM;
	}
	else
	{
		(this->*CommandTable[command & 31]) ();
	}
}

void FDC::SetData(uint8 d)
{
	// 󂯎󋵂`FbN
	if ((status & (S_RQM | S_DIO)) == S_RQM)
	{
		system->IntrReset();
		status &= ~S_RQM;
		data = d;
		switch (phase)
		{
		case IdlePhase:
			SetDataIdle();
			break;
		case CommandPhase:
			SetDataCommand();
			break;
		case ExecutePhase:
			SetDataExecute();
			break;
		case ExecWritePhase:
			SetDataExecWrite();
			break;
		case ErrorPhase:
			(this->*CommandTable[command & 31])();
			break;
		}
	}
}

// ---------------------------------------------------------------------------
//	FDC::GetData
//	FDC -> CPU
// ---------------------------------------------------------------------------

inline void FDC::GetDataResult()
{
	data = *bufptr++;
	Log1(" %.2x", data);
	if (--count)
	{
		status |= S_RQM;
	}
	else
	{
		ShiftToIdlePhase();
		Log0(" }\n");
	}
}

inline void FDC::GetDataExecute()
{
	(this->*CommandTable[command & 31])();
	if (phase == ExecReadPhase)
		GetDataExecRead();
}

inline void FDC::GetDataExecRead()
{
	data = *bufptr++;
	if (--count)
	{
		status |= S_RQM;
		system->Intr();
	}
	else
	{
		(this->*CommandTable[command & 31]) ();
	}
}

uint8 FDC::GetData()
{
	if ((status & (S_RQM | S_DIO)) == (S_RQM | S_DIO))
	{
		system->IntrReset();
		status &= ~S_RQM;
		switch (phase)
		{
		case ResultPhase:
			GetDataResult();
			break;
		case ExecutePhase:
			GetDataExecute();
			break;
		case ExecReadPhase:
			GetDataExecRead();
			break;
		case ErrorPhase:
			(this->*CommandTable[command & 31])();
			break;
		}
	}
	return data;
}

// ---------------------------------------------------------------------------

void FDC::TC()
{
	tcin = true;
	Log0("TC!\n");
	tcprevphase = phase;
	phase = TCPhase;
	(this->*CommandTable[command & 31]) ();
}

// ---------------------------------------------------------------------------
//	10	Invalid,		ScanEqual,		Invalid,		Invalid,
//	18	Invalid,		ScanLowOrEqual,	Invalid,		Invalid,
//	1c	Invalid,		ScanHighOrEqual,Invalid,		Invalid,

const FDC::CommandFunc FDC::CommandTable[32] =
{
	&FDC::CmdInvalid,				&FDC::CmdInvalid,			// 0
	&FDC::CmdReadDiagnostic,		&FDC::CmdSpecify,
	&FDC::CmdSenceDeviceStatus,		&FDC::CmdWriteData,			// 4
	&FDC::CmdReadData,				&FDC::CmdRecalibrate,
	&FDC::CmdSenceIntStatus,		&FDC::CmdWriteDeletedData,	// 8
	&FDC::CmdReadID,				&FDC::CmdInvalid,
	&FDC::CmdReadDeletedData,		&FDC::CmdWriteID,			// c
	&FDC::CmdInvalid,				&FDC::CmdSeek,
	&FDC::CmdInvalid,				&FDC::CmdInvalid,			// 10
	&FDC::CmdInvalid,				&FDC::CmdInvalid,			
	&FDC::CmdInvalid,				&FDC::CmdInvalid,			// 14
	&FDC::CmdInvalid,				&FDC::CmdInvalid,
	&FDC::CmdInvalid,				&FDC::CmdInvalid,			// 18
	&FDC::CmdInvalid,				&FDC::CmdInvalid,
	&FDC::CmdInvalid,				&FDC::CmdInvalid,			// 1c
	&FDC::CmdInvalid,				&FDC::CmdInvalid,
};

// ---------------------------------------------------------------------------
//	WriteID
// ---------------------------------------------------------------------------

void FDC::CmdWriteID()
{
	switch (phase)
	{
	case IdlePhase:
		Log0("WriteID ");
		ShiftToCommandPhase(5);
		return;
	
	case CommandPhase:
		Log4("(%.2x %.2x %.2x %.2x ", buffer[0], buffer[1], buffer[2], buffer[3]);
		Log1("%.2x)  ", buffer[4]);
		
		result = error = Result(0);
		
		wid.idr = 0;
		hdu     = buffer[0];
		wid.n   = buffer[1];
		wid.sc  = buffer[2];
		wid.gpl = buffer[3];
		wid.d   = buffer[4];
		if (wid.sc)
		{
			ShiftToExecWritePhase(4 * wid.sc);
			system->Intr();
			return;
		}
		bufptr = buffer;

	case TCPhase:
		if (wid.idr)
		{
			ShiftToResultPhase7();
			return;
		}

	case ExecWritePhase:
		wid.idr = (ID*) buffer;
		wid.sc = (bufptr - buffer) / 4;
		Log1("sc:%d ", wid.sc);
		
		WriteID();
		return;
	}

}

// ---------------------------------------------------------------------------
//	WriteID Execution
// ---------------------------------------------------------------------------

void FDC::WriteID()
{
#ifdef FDC_LOGGING
	Log2("\tWriteID  sc:%.2x N:%.2x\n", wid.sc, wid.n);
	for (int i=0; i<wid.sc; i++)
	{
		Log4("\t%.2x %.2x %.2x %.2x\n", 
			wid.idr[i].c, wid.idr[i].h, wid.idr[i].r, wid.idr[i].n); 
	}
#endif

	if ((hdu & 3) >= NUM_FLOPPY_DRIVES)
		result = Result(ST0_AT | ST0_NR);
	else 
	{
		uint8 flags = ((hdu >> 2) & 1) | (command & 0x40);
		result = fdu[hdu & 3]->WriteID(flags, wid);
	}

	idr.n = wid.n;
	ShiftToResultPhase7();
}

// ---------------------------------------------------------------------------
//	ReadID
// ---------------------------------------------------------------------------

void FDC::CmdReadID()
{
	switch (phase)
	{
	case IdlePhase:
		Log0("ReadID ");
		ShiftToCommandPhase(1);
		return;
	case CommandPhase:
		Log1("(%.2x)", buffer[0]);
		hdu = buffer[0];
		result = Result(0);
		
		ReadID();
		return;
	case TCPhase:
		phase = tcprevphase;
		return;
	}
}

// ---------------------------------------------------------------------------
//	ReadID Execution
// ---------------------------------------------------------------------------

void FDC::ReadID()
{
	if ((hdu & 3) >= NUM_FLOPPY_DRIVES)
		result = Result(ST0_AT | ST0_NR);
	else 
	{
		uint8 flags = ((hdu >> 2) & 1) | (command & 0x40);
		result = fdu[hdu & 3]->ReadID(flags, &idr);
	}

	ShiftToResultPhase7();
}

// ---------------------------------------------------------------------------
//	WriteDeletedData
// ---------------------------------------------------------------------------

void FDC::CmdWriteDeletedData()
{
	switch (phase)
	{
	case IdlePhase:
		Log0("WriteDeletedData ");
		ShiftToCommandPhase(8);
		return;
	
	case CommandPhase:
		GetSectorParameters();
		hdunext = hdu;
		result = error = Result(0);
		lastsect = false;

		ShiftToExecutePhase(0);
		return;
	
	case ExecutePhase:
		if (error)
		{
			result = error;
		}
		else if ((hdu & 3) >= NUM_FLOPPY_DRIVES || !(GetDeviceStatus(hdu & 3) & 0x20))
		{
			result = Result(ST0_AT | ST0_NR);
		} 
		else if (lastsect)
		{
			result = Result(ST0_AT | ST1_EN);
		}
		else
		{
			hdu = hdunext;
			ShiftToExecWritePhase(0x80 << Min(idr.n, 8));
			return;
		}

		system->Intr();
		status = S_CB;
		return;
	
	case ExecWritePhase:
		status &= ~(S_RQM | S_NDM);
		phase = ExecutePhase;
		WriteData(false, true);
		return;
	
	case TCPhase:
		Log1("\tTC at 0x%x byte\n", bufptr - buffer);
		if (tcprevphase == ExecWritePhase)
		{
			memset(bufptr, 0, count);
			Log0("flush");
			WriteData(true, true);
			return;
		}
		if (tcprevphase != ResultPhase)
		{
			ShiftToResultPhase7();
			return;
		}

		system->Intr();
		phase = tcprevphase;
		return;
	}
}

// ---------------------------------------------------------------------------
//	WriteData
// ---------------------------------------------------------------------------

void FDC::CmdWriteData()
{
	switch (phase)
	{
	case IdlePhase:
		Log0("WriteData ");
		ShiftToCommandPhase(8);
		return;
	
	case CommandPhase:
		GetSectorParameters();
		hdunext = hdu;
		result = error = Result(0);
		lastsect = false;

		ShiftToExecutePhase(0);
		return;
	
	case ExecutePhase:
		if (error)
		{
			result = error;
		}
		else if ((hdu & 3) >= NUM_FLOPPY_DRIVES || !(GetDeviceStatus(hdu & 3) & 0x20))
		{
			result = Result(ST0_AT | ST0_NR);
		} 
		else if (lastsect)
		{
			result = Result(ST0_AT | ST1_EN);
		}
		else
		{
			hdu = hdunext;
			ShiftToExecWritePhase(0x80 << Min(8, idr.n));
			return;
		}

		system->Intr();
		status = S_CB;
		return;
	
	case ExecWritePhase:
		status &= ~(S_RQM | S_NDM);
		phase = ExecutePhase;
		WriteData(false, false);
		return;
	
	case TCPhase:
		Log1("\tTC at 0x%x byte\n", bufptr - buffer);
		if (tcprevphase == ExecWritePhase)
		{
			memset(bufptr, 0, count);
			Log0("flush");
			WriteData(true, false);
			return;
		}
		if (tcprevphase != ResultPhase)
		{
			ShiftToResultPhase7();
			return;
		}

		system->Intr();
		phase = tcprevphase;
		return;
	}
}

void FDC::WriteData(bool last, bool deleted)
{
	Log4("\twrite %.2x %.2x %.2x %.2x\n", idr.c, idr.h, idr.r, idr.n);
	
	if (!(GetDeviceStatus(hdu & 3) & 0x20))
	{
		result = Result(ST0_AT | ST0_NR);
	} 
	else
	{
		uint8 flags = ((hdu >> 2) & 1) | (command & 0x40);
		result = fdu[hdu & 3]->WriteSector(flags, idr, buffer, deleted);
		if (!result)
		{
			lastsect = IDIncrement(&idr);
			if (!last)
				ShiftToExecutePhase(0);
			else
				ShiftToResultPhase7();
			return;
		}
	}
	system->Intr();
	status = S_CB;
	return;
}

// ---------------------------------------------------------------------------

void FDC::GetSectorParameters()
{
	Log4("(%.2x %.2x %.2x %.2x", buffer[0], buffer[1], buffer[2], buffer[3]);
	Log4(" %.2x %.2x %.2x %.2x)\n", buffer[4], buffer[5], buffer[6], buffer[7]);
	hdu   = buffer[0];
	idr.c = buffer[1];
	idr.h = buffer[2];
	idr.r = buffer[3];
	idr.n = buffer[4];
	eot   = buffer[5];
	dtl   = buffer[6];
}

// ---------------------------------------------------------------------------
//	ReadDiagnostic
// ---------------------------------------------------------------------------

void FDC::CmdReadDiagnostic()
{
	switch (phase)
	{
	case IdlePhase:
		Log0("ReadDiag ");
		command &= 0x7f;
		ShiftToCommandPhase(8);
		return;

	case CommandPhase:
		GetSectorParameters();
		result = error = Result(0);
		lastsect = false;
		status &= ~(S_RQM | S_NDM);
		ReadDiag(true);
		return; 

	case ExecReadPhase:
		status &= ~(S_RQM | S_NDM);
		ReadDiag(true);
		return;

	case ExecutePhase:
		Log0("validate\n");
		lastsect = IDIncrement(&idr);
		ShiftToExecReadPhase(0x80 << Min(8, idr.n));
		return;
	
	case TCPhase:
		Log1("\tTC at 0x%x byte\n", bufptr - buffer);
		if (tcprevphase != ResultPhase)
		{
			ShiftToResultPhase7();
			return;
		}
		system->Intr();
		phase = tcprevphase;
		return;

	case ErrorPhase:
		Log0("!\n");
		result = error;
		system->Intr();
		status = S_CB;
		return;
	}
}

void FDC::ReadDiag(bool first)
{
	if (!tcin)
	{
		Log4("\tpreread %.2x %.2x %.2x %.2x  ", idr.c, idr.h, idr.r, idr.n);
		
		if (lastsect)
		{
			error = Result(ST0_AT | ST1_EN);
		}
		else if ((hdu & 3) >= NUM_FLOPPY_DRIVES)
		{
			error = Result(ST0_AT | ST0_NR);
		}
		// ۂ̓fBXN}܂ő҂ׂH
		else if (!(GetDeviceStatus(hdu & 3) & 0x8) || !fdu[hdu & 3])
		{
			error = Result(ST0_AT | ST1_MA);
		}
		else
		{
			uint8 flags = ((hdu >> 2) & 1) | (command & 0x40);
			error = fdu[hdu & 3]->ReadDiagnostic(flags, idr, buffer, first);
			if (!error)
			{
				ShiftToExecutePhase(S_DIO);
				return;
			}
		}
		Log0("error? ");
		ShiftToErrorPhase(S_DIO);
	}
}

// ---------------------------------------------------------------------------
//	ReadDeletedData
// ---------------------------------------------------------------------------

void FDC::CmdReadDeletedData()
{
	switch (phase)
	{
	case IdlePhase:
		Log0("ReadDeletedData ");
		ShiftToCommandPhase(8);
		return;

	case CommandPhase:
		GetSectorParameters();
		hdunext = hdu;
		result = error = Result(0);
		lastsect = false;
		// 
	case ExecReadPhase:
		if (error & ST2_CM)
		{
			result = error;
			ShiftToResultPhase7();
			return;
		}
		status &= ~(S_RQM | S_NDM);
		ReadData(true);
		return;

	case ExecutePhase:
		Log0("validate\n");
		hdu = hdunext;
		lastsect = IDIncrement(&idr);
		ShiftToExecReadPhase(0x80 << Min(8, idr.n));
		return;
	
	case TCPhase:
		Log1("\tTC at 0x%x byte\n", bufptr - buffer);
		if (tcprevphase != ResultPhase)
		{
			ShiftToResultPhase7();
			return;
		}
		system->Intr();
		phase = tcprevphase;
		return;

	case ErrorPhase:
		Log0("!\n");
		result = error;
		system->Intr();
		status = S_CB;
		return;
	}
}

// ---------------------------------------------------------------------------
//	ReadData
// ---------------------------------------------------------------------------

void FDC::CmdReadData()
{
	switch (phase)
	{
	case IdlePhase:
		Log0("ReadData ");
		ShiftToCommandPhase(8);
		return;

	case CommandPhase:
		GetSectorParameters();
		hdunext = hdu;
		result = error = Result(0);
		lastsect = false;
		// 
	case ExecReadPhase:
		if (error & ST2_CM)
		{
			result = error;
			ShiftToResultPhase7();
			return;
		}
		status &= ~(S_RQM | S_NDM);
		ReadData(false);
		return;

	case ExecutePhase:
		Log0("validate\n");
		hdu = hdunext;
		lastsect = IDIncrement(&idr);
		ShiftToExecReadPhase(0x80 << Min(8, idr.n));
		return;
	
	case TCPhase:
		Log1("\tTC at 0x%x byte\n", bufptr - buffer);
		if (tcprevphase != ResultPhase)
		{
			ShiftToResultPhase7();
			return;
		}
		system->Intr();
		phase = tcprevphase;
		return;

	case ErrorPhase:
		Log0("!\n");
		result = error;
		system->Intr();
		status = S_CB;
		return;
	}
}

void FDC::ReadData(bool deleted)
{
	if (!tcin)
	{
		Log4("\tpreread %.2x %.2x %.2x %.2x  ", idr.c, idr.h, idr.r, idr.n);
		
		if (lastsect)
		{
			error = Result(ST0_AT | ST1_EN);
		}
		else if ((hdunext & 3) >= NUM_FLOPPY_DRIVES)
		{
			error = Result(ST0_AT | ST0_NR);
		}
		// ۂ̓fBXN}܂ő҂ׂH
		else if (!(GetDeviceStatus(hdunext & 3) & 0x8) || !fdu[hdunext & 3])
		{
			error = Result(ST0_AT | ST1_MA);
		}
		else
		{
			uint8 flags = ((hdunext >> 2) & 1) | (command & 0x40);
			error = fdu[hdunext & 3]->ReadSector(flags, idr, buffer);
			if (deleted)
				error = Result(error ^ ST2_CM);
			if (!(error & ~ST2_CM))
			{
				if (error & ST2_CM)
				{
					if (command & 0x20)
					{
						lastsect = IDIncrement(&idr);
						ReadData(deleted);
						return;
					}
				}
				ShiftToExecutePhase(S_DIO);
				return;
			}
		}
		Log0("error? ");
		ShiftToErrorPhase(S_DIO);
	}
}

// ---------------------------------------------------------------------------
//	command  EOT QlɃR[h
// ---------------------------------------------------------------------------

bool FDC::IDIncrement(FDU::ID* id)
{
	if (id->r++ < eot)
		return false;
	id->r = 1;
	if (command & 0x80)
	{
		id->h ^= 1;
		hdunext = hdu ^ 4;
		if (id->h & 1)
			return false;
	}
	id->c++;
	return true;
}

// ---------------------------------------------------------------------------
//	SenceIntState
// ---------------------------------------------------------------------------

void FDC::CmdSenceIntStatus()
{
	switch (phase)
	{
	case IdlePhase:
		if (int_requested)
		{
			Log0("SenceIntStatus\n");
			int_requested = true;
			buffer[0] = uint8(result);
			buffer[1] = pcn;
	
			ShiftToResultPhase(2);
		}
		else
		{
			Log0("Invalid(SenceIntStatus)\n");
			buffer[0] = uint8(ST0_IC);
			ShiftToResultPhase(1);
		}
		break;
	}
}

// ---------------------------------------------------------------------------
//	SenceDeviceStatus
// ---------------------------------------------------------------------------

void FDC::CmdSenceDeviceStatus()
{
	if (phase == IdlePhase)
	{
		Log0("SenceDeviceStatus ");
		ShiftToCommandPhase(1);
		return;
	}
	if (phase == CommandPhase)
	{
		Log1("(%.2x)\n", buffer[0]);
		buffer[0] = GetDeviceStatus(buffer[0] & 3);
		ShiftToResultPhase(1);
	}
}

uint8 FDC::GetDeviceStatus(uint dr)
{
	if (dr < NUM_FLOPPY_DRIVES)
		return fdu[dr]->SenceDeviceStatus() | dr;
	else
		return 0x80 | dr;
}

// ---------------------------------------------------------------------------
//	Recalibrate
// ---------------------------------------------------------------------------

void FDC::CmdRecalibrate()
{
	switch (phase)
	{
	case IdlePhase:
		Log0("Recalibrate ");
		ShiftToCommandPhase(1);
		break;
	
	case CommandPhase:
		Log1("(%.2x)  ", buffer[0]);
		ShiftToIdlePhase();

		seekstate |= 1 << (buffer[0] & 3);
		Seek(buffer[0] & 3, 0);
		break;
	}
}

// ---------------------------------------------------------------------------
//	Seek
// ---------------------------------------------------------------------------

void FDC::CmdSeek()
{
	switch (phase)
	{
	case IdlePhase:
		Log0("Seek ");
		ShiftToCommandPhase(2);
		break;
	
	case CommandPhase:
		Log2("(%.2x %.2x)  ", buffer[0], buffer[1]);
		ShiftToIdlePhase();

		seekstate |= 1 << (buffer[0] & 3);
		Seek(buffer[0] & 3, buffer[1]);
		break;
	}
}

// ---------------------------------------------------------------------------
//	w̃hCuV[N
// ---------------------------------------------------------------------------

void FDC::Seek(uint dr, uint cy)
{
	uint8 ds = GetDeviceStatus(dr);
	if (ds & ST3_RY)
	{
		fdu[dr]->Seek(cy);
		pcn = cy;
		result = Result((dr & 3) | ST0_SE);
		Log0("success.\n");
	}
	else
	{
		result = Result((dr & 3) | ST0_SE | ST0_NR | ST0_AT);
		Log0("error\n");
	}
	system->Intr();
	int_requested = true;
	seekstate &= ~(1 << (buffer[0] & 3));
	return;
}

// ---------------------------------------------------------------------------
//	Specify
// ---------------------------------------------------------------------------

void FDC::CmdSpecify()
{
	switch (phase)
	{
	case IdlePhase:
		Log0("Specify ");
		ShiftToCommandPhase(2);
		break;
	case CommandPhase:
		Log2("(%.2x %.2x)\n", buffer[0], buffer[1]);
		ShiftToIdlePhase();
		break;
	}
}

// ---------------------------------------------------------------------------
//	Invalid
// ---------------------------------------------------------------------------

void FDC::CmdInvalid()
{
	Log0("Invalid\n");
	buffer[0] = uint8(ST0_IC);
	ShiftToResultPhase(1);
}

// ---------------------------------------------------------------------------
//	Debug
// ---------------------------------------------------------------------------

void FDC::Txt(char* x)
{
	Log0(x);
}