// ---------------------------------------------------------------------------
//  M88 - PC88 emulator
//	Copyright (C) cisc 1998.
// ---------------------------------------------------------------------------
//	Config Dialog Box for M88
//	$Id: M88Config.cpp,v 1.3 1999/01/12 14:17:30 cisc Exp $

#define STRICT
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <stdlib.h>
#include <commctrl.h>
#include "resource.h"
#include "M88Config.h"
#include "misc.h"

static const char* AppName = "M88 for Windows";

// ---------------------------------------------------------------------------
//	Config::LoadEntry
// ---------------------------------------------------------------------------
	
bool Config::LoadEntry
(const char* inifile, const char* entry, int* value, int def, bool applydefault)
{
	int n = GetPrivateProfileInt(AppName, entry, -1, inifile);

	if (n == -1 && applydefault)
		n = def;
	if (n != -1)
	{
		*value = n;
		return true;
	}
	else
		return false;
}

// ---------------------------------------------------------------------------
//	M88Config::LoadDefaultDirectory
// ---------------------------------------------------------------------------

void Config::LoadDefaultDirectory(const char* inifile)
{
	if (flags & savedirectory)
	{
		char path[_MAX_PATH];
		if (GetPrivateProfileString(AppName, "Directory", ";", 
									path, _MAX_PATH, inifile))
		{
			if (path[0] != ';')
				SetCurrentDirectory(path);
		}
	}
}

// ---------------------------------------------------------------------------
//	M88Config::LoadConfig
// ---------------------------------------------------------------------------

#define VOLUME_BIAS 100

#define LOADVOLUMEENTRY(key, vol) \
	if (LoadEntry(inifile, key, &n, VOLUME_BIAS, applydefault)) \
	vol = n - VOLUME_BIAS;

void Config::LoadConfig(const char* inifile, bool applydefault)
{
	int n;

	n = subcpucontrol | savedirectory;
	LoadEntry(inifile, "Flags", &flags, n, applydefault);
	
	if (LoadEntry(inifile, "CPUClock", &n, 4, applydefault))
		clock = Limit(n, 100, 1);

	if (LoadEntry(inifile, "Speed", &n, 1000, applydefault))
		speed = Limit(n, 2000, 500);

	if (LoadEntry(inifile, "RefreshTiming", &n, 3, applydefault))
		refreshtiming = Limit(n, 4, 1);

	if (LoadEntry(inifile, "BASICMode", &n, PC88::N88V2, applydefault))
	{
		if (   n == PC88::N80    || n == PC88::N88V1 
			|| n == PC88::N88V1H || n == PC88::N88V2)
			basicmode = PC88::BASICMODE(n);
		else
			basicmode = PC88::N88V2;
	}

	if (LoadEntry(inifile, "Sound", &n, 2, applydefault))
		sound = Limit(n, 5, 0);
	
	if (LoadEntry(inifile, "OPNClock", &n, 3993600, applydefault))
		opnclock = Limit(n, 10000000, 1000000);
	
	if (LoadEntry(inifile, "ERAMBank", &n, 4, applydefault))
		nerams = Limit(n, 256, 4);

	if (LoadEntry(inifile, "KeyboardType", &n, GetKeyboardType(1), applydefault))
		keytype = n;

	if (LoadEntry(inifile, "Switches", &n, 1829, applydefault))
		dipsw = n;

	if (LoadEntry(inifile, "SoundBuffer", &n, 200, applydefault))
		soundbuffer = Limit(n, 1000, 100);

	if (LoadEntry(inifile, "PadSensibility", &n, 16384, applydefault))
		padsensibility = Limit(n, 32767, 10);
	
	LOADVOLUMEENTRY("VolumeFM", volfm);
	LOADVOLUMEENTRY("VolumeSSG", volssg);
	LOADVOLUMEENTRY("VolumeADPCM", voladpcm);
	LOADVOLUMEENTRY("VolumeRhythm", volrhythm);
	LOADVOLUMEENTRY("VolumeBD", volbd);
	LOADVOLUMEENTRY("VolumeSD", volsd);
	LOADVOLUMEENTRY("VolumeTOP", voltop);
	LOADVOLUMEENTRY("VolumeHH", volhh);
	LOADVOLUMEENTRY("VolumeTOM", voltom);
	LOADVOLUMEENTRY("VolumeRIM", volrim);

}

// ---------------------------------------------------------------------------
//	Config::SaveEntry
// ---------------------------------------------------------------------------

bool Config::SaveEntry
(const char* inifile, const char* entry, int value, bool applydefault)
{
	char buf[_MAX_PATH];
	if (applydefault || -1 != GetPrivateProfileInt(AppName, entry, -1, inifile))
	{
		wsprintf(buf, "%d", value);
		WritePrivateProfileString(AppName, entry, buf, inifile);
		return true;
	}
	else
		return false;
}

bool Config::SaveEntry
(const char* inifile, const char* entry, const char* value, bool applydefault)
{
	bool apply = applydefault;
	if (!applydefault)
	{
		char buf[8];
		GetPrivateProfileString(AppName, entry, ";", buf, sizeof(buf), inifile);
		if (buf[0] != ';')
			apply = true;
	}
	if (apply)
		WritePrivateProfileString(AppName, entry, value, inifile);
	return apply;
}

// ---------------------------------------------------------------------------
//	Config::SaveConfig
// ---------------------------------------------------------------------------

void Config::SaveConfig(const char* inifile, bool writedefault)
{
	char buf[_MAX_PATH];
	GetCurrentDirectory(_MAX_PATH, buf);
	SaveEntry(inifile, "Directory", buf, writedefault);

	SaveEntry(inifile, "Flags", flags, writedefault);
	SaveEntry(inifile, "CPUClock", clock, writedefault);
	SaveEntry(inifile, "Speed", speed, writedefault);
	SaveEntry(inifile, "RefreshTiming", refreshtiming, writedefault);
	SaveEntry(inifile, "BASICMode", basicmode, writedefault);
	SaveEntry(inifile, "Sound", sound, writedefault);
	SaveEntry(inifile, "Switches", dipsw, writedefault);
	SaveEntry(inifile, "SoundBuffer", soundbuffer, writedefault);
	SaveEntry(inifile, "PadSensibility", padsensibility, writedefault);

	SaveEntry(inifile, "VolumeFM", volfm + VOLUME_BIAS, writedefault);
	SaveEntry(inifile, "VolumeSSG", volssg + VOLUME_BIAS, writedefault);
	SaveEntry(inifile, "VolumeADPCM", voladpcm + VOLUME_BIAS, writedefault);
	SaveEntry(inifile, "VolumeRhythm", volrhythm + VOLUME_BIAS, writedefault);
	SaveEntry(inifile, "VolumeBD", volbd + VOLUME_BIAS, writedefault);
	SaveEntry(inifile, "VolumeSD", volsd + VOLUME_BIAS, writedefault);
	SaveEntry(inifile, "VolumeTOP", voltop + VOLUME_BIAS, writedefault);
	SaveEntry(inifile, "VolumeHH", volhh + VOLUME_BIAS, writedefault);
	SaveEntry(inifile, "VolumeTOM", voltom + VOLUME_BIAS, writedefault);
	SaveEntry(inifile, "VolumeRIM", volrim + VOLUME_BIAS, writedefault);
}
